"""Module containing decorators"""

__author__ = 'Robert Meyer'

import functools
import warnings
import logging
import time


def manual_run(turn_into_run=True, store_meta_data=True, clean_up=True):
    """Can be used to decorate a function as a manual run function.

    This can be helpful if you want the run functionality without using an environment.

    :param turn_into_run:

        If the trajectory should become a `single run` with more specialized functionality
        during a single run.

    :param store_meta_data:

        If meta-data like runtime should be automatically stored

    :param clean_up:

        If all data added during the single run should be removed, only works
        if ``turn_into_run=True``.

    """
    def wrapper(func):
        @functools.wraps(func)
        def new_func(traj, *args, **kwargs):
            do_wrap = not traj._run_by_environment
            if do_wrap:
                traj.f_start_run(turn_into_run=turn_into_run)
            result = func(traj, *args, **kwargs)
            if do_wrap:
                traj.f_finalize_run(store_meta_data=store_meta_data,
                                    clean_up=clean_up)
            return result

        return new_func

    return wrapper


def deprecated(msg=''):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    :param msg:

        Additional message added to the warning.

    """

    def wrapper(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warning_string = "Call to deprecated function or property `%s`." % func.__name__
            warning_string = warning_string + ' ' + msg
            warnings.warn(
                warning_string,
                category=DeprecationWarning,
            )
            return func(*args, **kwargs)

        return new_func

    return wrapper


def copydoc(fromfunc, sep="\n"):
    """Decorator: Copy the docstring of `fromfunc`

    If the doc contains a line with the keyword `ABSTRACT`,
    like `ABSTRACT: Needs to be defined in subclass`, this line and the line after are removed.

    """

    def _decorator(func):
        sourcedoc = fromfunc.__doc__

        # Remove the ABSTRACT line:
        split_doc = sourcedoc.split('\n')
        split_doc_no_abstract = [line for line in split_doc if not 'ABSTRACT' in line]

        # If the length is different we have found an ABSTRACT line
        # Finally we want to remove the final blank line, otherwise
        # we would have three blank lines at the end
        if len(split_doc) != len(split_doc_no_abstract):
            sourcedoc = '\n'.join(split_doc_no_abstract[:-1])

        if func.__doc__ is None:
            func.__doc__ = sourcedoc
        else:
            func.__doc__ = sep.join([sourcedoc, func.__doc__])
        return func

    return _decorator


def kwargs_mutual_exclusive(param1_name, param2_name, map2to1=None):
    """ If there exist mutually exclusive parameters checks for them and maps param2 to 1."""
    def wrapper(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            if param2_name in kwargs:
                if param1_name in kwargs:
                    raise ValueError('You cannot specify `%s` and `%s` at the same time, '
                                     'they are mutually exclusive.' % (param1_name, param2_name))
                param2 = kwargs.pop(param2_name)
                if map2to1 is not None:
                    param1 = map2to1(param2)
                else:
                    param1 = param2
                kwargs[param1_name] = param1

            return func(*args, **kwargs)

        return new_func

    return wrapper


def kwargs_api_change(old_name, new_name=None):
    """This is a decorator which can be used if a kwarg has changed
    its name over versions to also support the old argument name.

    Issues a warning if the old keyword argument is detected and
    converts call to new API.

    :param old_name:

        Old name of the keyword argument

    :param new_name:

        New name of keyword argument

    """

    def wrapper(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):

            if old_name in kwargs:
                if new_name is None:
                    warning_string = 'Using deprecated keyword argument `%s` in function `%s`. ' \
                                 'This keyword is no longer supported, please don`t use it ' \
                                 'anymore.' % (old_name, func.__name__)
                else:
                    warning_string = 'Using deprecated keyword argument `%s` in function `%s`, ' \
                                 'please use keyword `%s` instead.' % \
                                 (old_name, func.__name__, new_name)
                warnings.warn(warning_string, category=DeprecationWarning)
                value = kwargs.pop(old_name)
                if new_name is not None:
                    kwargs[new_name] = value

            return func(*args, **kwargs)

        return new_func

    return wrapper


def not_in_run(func):
    """This is a decorator that signaling that a function is not available during a single run.

    """
    doc = func.__doc__
    na_string = '''\nATTENTION: This function is not available during a single run!\n'''

    if doc is not None:
        func.__doc__ = '\n'.join([doc, na_string])
    func._not_in_run = True

    @functools.wraps(func)
    def new_func(self, *args, **kwargs):

        if self._is_run:
            raise TypeError('Function `%s` is not available during a single run.' %
                            func.__name__)

        return func(self, *args, **kwargs)

    return new_func


def with_open_store(func):
    """This is a decorator that signaling that a function is only available if the storage is open.

    """
    doc = func.__doc__
    na_string = '''\nATTENTION: This function can only be used if the store is open!\n'''

    if doc is not None:
        func.__doc__ = '\n'.join([doc, na_string])
    func._with_open_store = True

    @functools.wraps(func)
    def new_func(self, *args, **kwargs):

        if not self.traj.v_storage_service.is_open:
            raise TypeError('Function `%s` is only available if the storage is open.' %
                            func.__name__)

        return func(self, *args, **kwargs)

    return new_func


def retry(n, errors, wait=0.0, logger_name=None):
    """This is a decorator that retries a function.

    Tries `n` times and catches a given tuple of `errors`.

    If the `n` retries are not enough, the error is reraised.

    If desired `waits` some seconds.

    Optionally takes a 'logger_name' of a given logger to print the caught error.

    """

    def wrapper(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            retries = 0
            while True:
                try:
                    result = func(*args, **kwargs)
                    if retries and logger_name:
                        logger = logging.getLogger(logger_name)
                        logger.debug('Retry of `%s` successful' % func.__name__)
                    return result
                except errors:
                    if retries >= n:
                        if logger_name:
                            logger = logging.getLogger(logger_name)
                            logger.exception('I could not execute `%s` with args %s and kwargs %s, '
                                             'starting next try. ' % (func.__name__,
                                                                      str(args),
                                                                      str(kwargs)))
                        raise
                    elif logger_name:
                        logger = logging.getLogger(logger_name)
                        logger.debug('I could not execute `%s` with args %s and kwargs %s, '
                                         'starting next try. ' % (func.__name__,
                                                                  str(args),
                                                                  str(kwargs)))
                    retries += 1
                    if wait:
                        time.sleep(wait)
        return new_func

    return wrapper


def _prfx_getattr_(obj, item):
    """Replacement of __getattr__"""
    if item.startswith('f_') or item.startswith('v_'):
        return getattr(obj, item[2:])
    raise AttributeError('`%s` object has no attribute `%s`' % (obj.__class__.__name__, item))


def _prfx_setattr_(obj, item, value):
    """Replacement of __setattr__"""
    if item.startswith('v_'):
        return setattr(obj, item[2:], value)
    else:
        return super(obj.__class__, obj).__setattr__(item, value)


def prefix_naming(cls):
    """Decorate that adds the prefix naming scheme"""
    if hasattr(cls, '__getattr__'):
        raise TypeError('__getattr__ already defined')
    cls.__getattr__ = _prfx_getattr_
    cls.__setattr__ = _prfx_setattr_
    return cls




"""Module running all examples in the examples directory

Suppresses all openings of plots
"""
__author__ = 'Robert Meyer'

import glob
import os
import sys
import platform
import subprocess

try:
    import brian2
except ImportError:
    print('No BRIAN module found, will skip the example')
    brian2 = None

system = platform.system()
print('*** Running under %s ***' % str(system))

to_skip = set()
to_skip.add('21')

if brian2 is None:
    to_skip.add('23')
    to_skip.add('24')
if system == 'Windows':
    # Appveyor is too slow for this example, so we skip it
    to_skip.add('13')
    to_skip.add('19')
    to_skip.add('20')

if len(to_skip) == 0:
    print ('----- I will run all tests -----')
else:
    print('----- I will skip the following tests: `%s` ----' % to_skip)


def skip(name):
    for item in to_skip:
        if item in name:
            return True
    return False


def prepend_mpl_import(filename):
    """Writes a new python file and prepends the a matplotlib import.

    This avoids opening the plots within the test suite
    """
    with open(filename, mode='r') as fh:
        file_text = fh.read()

    import_matplotlib = ('import matplotlib\n'
                        'matplotlib.use(\'Agg\')\n'
                        'print(\'IMPORTED MATPLOTLIB WITH AGG\')\n')
    new_filename = 'tmp_' + filename
    with open(new_filename, mode='w') as fh:
        fh.write(import_matplotlib)
        fh.write(file_text)

    return new_filename


def execute_example(filename):
    """Executes a file as script.

    First prepends matplotlib import

    """
    #filename = os.path.join(os.getcwd(), filename)
    new_filename = None
    try:
        new_filename = prepend_mpl_import(filename)
        retcode = subprocess.call(sys.executable + ' ' + new_filename, shell=True)
        if retcode != 0:
            print('### ERROR: Example `%s` failed! ###' % filename)
            sys.exit(retcode)
        else:
            print('#### Example successful ####')
    finally:
        if new_filename is not None:
            os.remove(new_filename)


def main():
    os.chdir(os.path.join('..','..','examples'))
    sys.path.append(os.getcwd())
    simple_examples = glob.glob('*.py')
    assert len(simple_examples) == 22 - 4 + 1 + 2
    # + Number of Examples - Number of Packages +  __init__.py + 19b and 19c

    for simple_example in simple_examples:
        if simple_example == '__init__':
            continue

        if skip(simple_example):
            print("---------- Skipping %s ----------" % simple_example)
        else:
            print("########## Running %s ###########" % simple_example)
            execute_example(simple_example)

    ex17 = 'example_17_wrapping_an_existing_project'
    if skip(ex17):
        print("------- Skipping %s -------" % ex17)
    else:
        os.chdir(ex17)
        sys.path.append(os.getcwd())
        print('Running original')
        execute_example('original.py')
        print('Running pypet wrapping')
        execute_example('pypetwrap.py')
        os.chdir('..')

    ex13 = 'example_13_post_processing'
    if skip(ex13):
        print("------- Skipping %s -------" % ex13)
    else:
        print("########## Running %s ###########" % ex13)
        os.chdir(ex13)
        sys.path.append(os.getcwd())
        print("Running main")
        execute_example('main.py')
        print("Running analysis")
        execute_example('analysis.py')
        print("Running pipeline")
        execute_example('pipeline.py')
        os.chdir('..')

    ex24 = 'example_24_large_scale_brian2_simulation'
    if skip(ex24):
        print("------- Skipping %s -------" % ex24)
    else:
        print("########## Running %s ###########" % ex24)

        os.chdir(ex24)
        sys.path.append(os.getcwd())
        print("Running script")
        execute_example('runscript.py')
        print("Running analysis")
        execute_example('plotff.py')
        os.chdir('..')


if __name__ == '__main__':
    main()


__author__ = ['Henri Bunting', 'Robert Meyer']


import numpy as np

try:
    import brian2
    from brian2.monitors.spikemonitor import SpikeMonitor
    from brian2.monitors.statemonitor import StateMonitor
    from brian2.monitors.ratemonitor import PopulationRateMonitor
    from pypet.brian2.parameter import Brian2MonitorResult
except ImportError:
    brian2 = None

from pypet.tests.testutils.ioutils import unittest
from pypet.tests.unittests.parameter_test import ResultTest
from pypet.tests.unittests.brian2tests.run_a_brian2_network import run_network
import pypet.utils.comparisons as comp



@unittest.skipIf(brian2 is None, 'Can only be run with brian2!')
class Brian2MonitorTest(ResultTest):

    tags = 'unittest', 'brian2', 'result', 'monitor', 'henri'

    @classmethod
    def setUpClass(cls):
        if brian2 is not None:
            Brian2MonitorResult.monitors = run_network()

    def setUp(self):
        self.monitors = Brian2MonitorResult.monitors
        self.make_results()
        self.make_constructor()

    def make_constructor(self):
        self.Constructor=Brian2MonitorResult
        self.dynamic_imports = [Brian2MonitorResult]


    def check_spike_monitor(self, res, monitor):
        self.assertTrue(comp.nested_equal(monitor.num_spikes, res.num_spikes))
        self.assertTrue(comp.nested_equal(str(monitor.source), res.source))
        self.assertEqual(monitor.name,res.name)
        self.assertTrue(comp.nested_equal(monitor.t[:], res.t))
        self.assertTrue(comp.nested_equal(monitor.i[:], res.i))
        for idx, varname in enumerate(monitor.record_variables):
            self.assertTrue(comp.nested_equal(getattr(monitor, varname)[:], res.f_get(varname)))



    def check_state_monitor(self, res, monitor):
        self.assertEqual(monitor.record_variables, res.record_variables)
        self.assertEqual(monitor.name,res.name)

        times=np.array(monitor.t)
        if len(times)>0:
            self.assertTrue(comp.nested_equal(times, res.t))


        for idx, varname in enumerate(monitor.record_variables):
            self.assertTrue(comp.nested_equal(monitor.record, res.record))
            self.assertTrue(comp.nested_equal(monitor.when, res.when))
            self.assertTrue(comp.nested_equal(str(monitor.source), res.source))

            self.assertTrue(comp.nested_equal(getattr(monitor, varname), res.f_get(varname)))

    def check_population_rate_monitor(self, res, monitor):
        self.assertEqual(str(monitor.source),res.source)
        self.assertEqual(monitor.name,res.name)
        #self.assertTrue(comp.nested_equal(monitor._bin,res.bin))
        self.assertTrue(comp.nested_equal(monitor.rate[:],res.rate))
        self.assertTrue(comp.nested_equal(monitor.t[:],res.t))
        #self.assertTrue(comp.nested_equal(monitor.delay,res.delay))

    def test_failing_adding_another_monitor(self):

        for res in self.results.values():
            with self.assertRaises(TypeError):
                res.f_set(monitor=self.monitors.values()[0])


    def test_the_insertion_made_implicetly_in_setUp(self):

        for key, monitor in self.monitors.items():
            res = self.results[key]
            self.assertEqual(res.testtestextra, 42)
            self.assertEqual(res.v_monitor_type, monitor.__class__.__name__)


            if isinstance(monitor, SpikeMonitor):
                self.check_spike_monitor(res, monitor)

            elif isinstance(monitor, StateMonitor):
                self.check_state_monitor(res, monitor)

            elif  isinstance(monitor, PopulationRateMonitor):
                self.check_population_rate_monitor(res, monitor)

            else:
                raise ValueError('Monitor Type %s is not supported (yet)' % str(type(monitor)))

    def test_dir(self):
        res = self.results['SpikeMonitor']
        self.assertTrue('i' in res)
        self.assertTrue('i' in dir(res))

    def test_set_item_via_number(self):
        res = self.results['SpikeMonitor']
        res[0] = 'Hi'
        res[777] = 777

        self.assertTrue(getattr(res, res.v_name) == 'Hi')
        self.assertTrue(res.f_get(0) == 'Hi')
        self.assertTrue(getattr(res, res.v_name + '_777') == 777)
        self.assertTrue(res[777] == 777)
        self.assertTrue(res.f_get(777) == 777)

        self.assertTrue(0 in res)
        self.assertTrue(777 in res)
        self.assertTrue(99999999 not in res)

        del res[0]
        self.assertTrue(0 not in res)

        del res[777]
        self.assertTrue(777 not in res)

    def make_results(self):
        self.results={}
        for key,monitor in self.monitors.items():
            self.results[key]=Brian2MonitorResult(key,monitor)

            self.results[key].f_set(testtestextra=42)



__author__ = 'Robert Meyer'

import os

from pypet.tests.testutils.ioutils import  run_suite, make_temp_dir,\
    parse_args, handle_config_file
from pypet.tests.testutils.data import  TrajectoryComparator
from pypet import Environment
import pypet


class ConfigParseTest(TrajectoryComparator):

    tags = 'configparser', 'unittest'

    def setUp(self):

        pypet_path = os.path.abspath(os.path.dirname(pypet.environment.__file__))
        init_path = os.path.join(pypet_path, 'logging')
        self.config_file = os.path.join(init_path, 'env_config_test.ini')
        self.parser = handle_config_file(self.config_file)


    def test_parsing(self):

        filename = make_temp_dir('config_test.hdf5')
        env = Environment(filename=filename, config=self.parser)

        traj = env.v_traj
        self.assertTrue(traj.v_auto_load)
        self.assertEqual(traj.v_storage_service.filename, filename)

        self.assertEqual(traj.x, 42)
        self.assertEqual(traj.f_get('y').v_comment, 'This is the second variable')
        self.assertTrue(traj.testconfig)

        self.assertTrue(env._logging_manager.log_config is not None)
        self.assertTrue(env._logging_manager._sp_config is not None)

        env.f_disable_logging()


if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)


__author__ = ('Robert Meyer', 'Mehmet Nevvaf Timur')

import sys
import unittest

import os
import platform
import numpy as np
import pandas as pd
import tables as pt
from pypet import SharedPandasFrame,  ObjectTable,  make_ordinary_result, Result, \
    make_shared_result, compact_hdf5_file, SharedCArray, SharedEArray, \
    SharedVLArray

from pypet.tests.testutils.ioutils import get_root_logger, parse_args, run_suite
from pypet.tests.testutils.ioutils import make_temp_dir, make_trajectory_name, unittest
from pypet.tests.testutils.data import TrajectoryComparator
from pypet import Trajectory, SharedResult, SharedTable, SharedArray, load_trajectory, StorageContextManager


class MyTable(pt.IsDescription):

    id = pt.Int32Col()
    name = pt.StringCol(15)
    surname = pt.StringCol(15)
    weight = pt.FloatCol()


class StorageDataTrajectoryTests(TrajectoryComparator):

    tags = 'unittest', 'trajectory', 'shared', 'hdf5'

    def test_conversions(self):
        filename = make_temp_dir('hdf5manipulation.hdf5')
        traj = Trajectory(name=make_trajectory_name(self), filename=filename)

        trajname = traj.v_name
        traj.v_standard_result = SharedResult

        traj.f_store(only_init=True)

        traj.f_add_result('shared_data')

        thedata = np.zeros((1000, 1000))
        myarray = SharedArray('array', traj.shared_data, trajectory=traj)
        traj.shared_data['array'] = myarray
        mytable = SharedTable('t1', traj.shared_data, trajectory=traj)
        traj.shared_data['t1'] = mytable
        dadict = {'hi': [1, 2, 3, 4, 5], 'shu': ['bi', 'du', 'da', 'ha', 'hui']}
        dadict2 = {'answer': [42]}
        res = traj.f_add_result('shared.dfs')
        res['df'] = SharedPandasFrame()
        res['df'].create_shared_data(data=pd.DataFrame(dadict), trajectory=traj)
        frame = SharedPandasFrame('df1', traj.f_get('shared.dfs'), trajectory=traj,
                                  add_to_parent=True)
        frame.create_shared_data(data=pd.DataFrame(dadict2),)
        res['df1'] = frame

        traj.f_add_result('mylist', [1, 2, 3])
        traj.f_add_result('my.mytuple', k=(1, 2, 3), wa=42)
        traj.f_add_result('my.myarray', np.zeros((50, 50)))
        traj.f_add_result('my.myframe', data=pd.DataFrame(dadict2))
        traj.f_add_result('my.mytable', ObjectTable(data=dadict2))

        myarray.create_shared_data(data=thedata)
        mytable.create_shared_data(first_row={'hi': 'hi'.encode('utf-8'), 'huhu': np.ones(3)})

        traj.f_store()

        data = myarray.read()
        myarray.get_data_node()
        self.assertTrue(np.all(data == thedata))

        with StorageContextManager(traj):
            myarray[2, 2] = 10
            data = myarray.read()
            self.assertTrue(data[2, 2] == 10)

        self.assertTrue(data[2, 2] == 10)
        self.assertFalse(traj.v_storage_service.is_open)

        traj = load_trajectory(name=trajname, filename=filename, load_all=2,
                               dynamic_imports=SharedResult)

        make_ordinary_result(traj.shared_data, 'array', trajectory=traj)
        array = traj.shared_data.array
        self.assertTrue(isinstance(array, np.ndarray))
        thedata[2, 2] = 10
        self.assertTrue(np.all(array == thedata))

        make_ordinary_result(traj.shared_data, 't1', trajectory=traj,)
        t1 = traj.shared_data.t1
        self.assertTrue(isinstance(t1, ObjectTable))
        self.assertTrue(np.all(t1['huhu'][0] == np.ones(3)))

        dfs = traj.shared.dfs
        make_ordinary_result(traj.shared.dfs, 'df', trajectory=traj)
        theframe = dfs.f_get('df')
        self.assertTrue(isinstance(dfs, Result))
        self.assertTrue(isinstance(theframe, pd.DataFrame))
        self.assertTrue(theframe['hi'][0] == 1)

        listres = traj.f_get('mylist')
        listres = make_shared_result(listres, 0, trajectory=traj)
        with StorageContextManager(traj):
            self.assertTrue(listres[0][2] == 3)
            listres[0][0] = 4

        self.assertTrue(listres[0][0] == 4)
        listres = make_ordinary_result(listres, 0, trajectory=traj)
        traj = load_trajectory(name=trajname, filename=filename, load_all=2,
                               dynamic_imports=SharedResult)
        mylist = traj.mylist
        self.assertTrue(isinstance(listres, Result))
        self.assertTrue(mylist[0] == 4)
        self.assertTrue(isinstance(mylist, list))

        mytuple = traj.mytuple

        with self.assertRaises(AttributeError):
            mytuple = make_shared_result(mytuple, 'mylist', traj, new_class=SharedArray)

        mytuple = make_shared_result(mytuple, 'k', traj, new_class=SharedArray)
        self.assertTrue(mytuple.k[1] == 2)

        mytuple = make_ordinary_result(mytuple, 'k', trajectory=traj)
        self.assertTrue(isinstance(mytuple.k, tuple))
        self.assertTrue(mytuple.k[2] == 3)

        myframe = traj.myframe
        myframe = make_shared_result(myframe, 'data', traj)

        theframe = myframe.data.read()
        self.assertTrue(theframe['answer'][0] == 42)

        myframe = make_ordinary_result(myframe, 'data', trajectory=traj)
        traj.f_load_item(myframe)
        self.assertTrue(myframe.data['answer'][0] == 42)

        mytable = traj.f_get('mytable')
        mytable = make_shared_result(mytable, 0, traj)

        self.assertTrue(isinstance(mytable[0], SharedTable))
        rows = mytable.mytable.read()

        self.assertTrue(rows[0][0] == 42)

        mytable = make_ordinary_result(mytable, 0, trajectory=traj)

        self.assertTrue(isinstance(mytable, Result))
        self.assertTrue(mytable[0]['answer'][0] == 42)

    def test_storing_and_manipulating(self):
        filename = make_temp_dir('hdf5manipulation.hdf5')
        traj = Trajectory(name=make_trajectory_name(self), filename=filename)
        trajname = traj.v_name

        thedata = np.zeros((1000, 1000))
        res = traj.f_add_result(SharedResult, 'shared')
        myarray = SharedArray('array', res, trajectory=traj, add_to_parent=True)
        mytable = SharedTable('t1', res, trajectory=traj, add_to_parent=True)
        mytable2 = SharedTable('t2', res, trajectory=traj, add_to_parent=True)
        mytable3 = SharedTable('t3', res, trajectory=traj, add_to_parent=True)

        traj.f_store(only_init=True)
        myarray.create_shared_data(data=thedata)
        mytable.create_shared_data(first_row={'hi': 'hi'.encode('utf-8'), 'huhu': np.ones(3)})
        mytable2.create_shared_data(description={'ha': pt.StringCol(2, pos=0), 'haha': pt.FloatCol(pos=1)})
        mytable3.create_shared_data(description={'ha': pt.StringCol(2, pos=0), 'haha': pt.FloatCol(pos=1)})

        traj.f_store()

        newrow = {'ha': 'hu', 'haha': 4.0}

        with self.assertRaises(TypeError):
            traj.shared.t2.row

        with StorageContextManager(traj) as cm:
            row = traj.shared.t2.row
            for irun in range(11):
                for key, val in newrow.items():
                    row[key] = val
                row.append()
            traj.shared.t3.flush()

        data = myarray.read()
        myarray.get_data_node()
        self.assertTrue(np.all(data == thedata))

        with StorageContextManager(traj):
            myarray[2, 2] = 10
            data = myarray.read()
            self.assertTrue(data[2, 2] == 10)

        self.assertTrue(data[2, 2] == 10)
        self.assertFalse(traj.v_storage_service.is_open)

        traj = load_trajectory(name=trajname, filename=filename)

        traj.f_load(load_data=2)

        traj.shared.t2.traj = traj
        traj.shared.t1.traj = traj
        traj.shared.array.traj = traj

        self.assertTrue(traj.shared.t2.nrows == 11, '%s != 11' % str(traj.shared.t2.nrows))
        self.assertTrue(traj.shared.t2[0]['ha'] == 'hu'.encode('utf-8'), traj.shared.t2[0]['ha'])
        self.assertTrue(traj.shared.t2[1]['ha'] == 'hu'.encode('utf-8'), traj.shared.t2[1]['ha'])
        self.assertTrue('huhu' in traj.shared.t1.colnames)
        self.assertTrue(traj.shared.array[2, 2] == 10)

    @unittest.skipIf(platform.system() == 'Windows', 'Not supported under Windows')
    def test_compacting(self):
        filename = make_temp_dir('hdf5compacting.hdf5')
        traj = Trajectory(name=make_trajectory_name(self), filename=filename)
        trajname = traj.v_name
        traj.v_storage_service.complevel = 7

        first_row = {'ha': 'hi'.encode('utf-8'), 'haha': np.zeros((3, 3))}

        traj.f_store(only_init=True)

        traj.f_add_result('My.Tree.Will.Be.Deleted', 42)
        traj.f_add_result('Mine.Too.HomeBoy', 42, comment='Don`t cry for me!')

        res = traj.f_add_result(SharedResult, 'myres')

        res['myres'] = SharedTable()

        res['myres'].create_shared_data(first_row=first_row)

        with StorageContextManager(traj):
            traj.myres
            for irun in range(10000):
                row = traj.myres.row
                for key in first_row:
                    row[key] = first_row[key]
                row.append()
        traj.f_store()
        del traj
        traj = load_trajectory(name=trajname, filename=filename, load_all=2)
        with StorageContextManager(traj) as cm:
            tb = traj.myres.get_data_node()
            tb.remove_rows(1000, 10000)

            cm.flush_store()
            self.assertTrue(traj.myres.nrows == 1001)

        traj.f_delete_item(traj.My, recursive=True)
        traj.f_delete_item(traj.Mine, recursive=True)

        size = os.path.getsize(filename)
        get_root_logger().info('Filesize is %s' % str(size))
        name_wo_ext, ext = os.path.splitext(filename)
        backup_file_name = name_wo_ext + '_backup' + ext
        code = compact_hdf5_file(filename, keep_backup=True)
        if code != 0:
            raise RuntimeError('ptrepack fail')
        backup_size = os.path.getsize(backup_file_name)
        self.assertTrue(backup_size == size)
        new_size = os.path.getsize(filename)
        get_root_logger().info('New filesize is %s' % str(new_size))
        self.assertTrue(new_size < size, "%s > %s" % (str(new_size), str(size)))

    def test_all_arrays(self):
        filename = make_temp_dir('hdf5arrays.hdf5')
        traj = Trajectory(name=make_trajectory_name(self), filename=filename)
        trajname = traj.v_name

        npearray = np.ones((2, 10, 3), dtype=np.float)
        thevlarray = np.array(['j'.encode('utf-8'), 22.2, 'gutter'.encode('utf-8')])
        traj.f_store(only_init=True)
        res = traj.f_add_result(SharedResult, 'arrays')
        res['carray'] = SharedCArray()
        res['carray'].create_shared_data(shape=(10, 10), atom=pt.atom.FloatAtom())
        res['earray'] = SharedEArray()
        res['earray'].create_shared_data(obj=npearray)
        res['vlarray'] = SharedVLArray()
        res['vlarray'].create_shared_data(obj=thevlarray)
        res['array'] = SharedArray()
        res['array'].create_shared_data(data=npearray)

        traj.f_store()

        traj = load_trajectory(name=trajname, filename=filename, load_all=2,
                               dynamic_imports=SharedResult)

        toappned = [44, 'k'.encode('utf-8')]
        with StorageContextManager(traj):
            a1 = traj.arrays.array
            a1[0, 0, 0] = 4.0

            a2 = traj.arrays.carray
            a2[0, 1] = 4

            a4 = traj.arrays.vlarray
            a4.append(toappned)

            a3 = traj.arrays.earray
            a3.append(np.zeros((1, 10, 3)))

        traj = load_trajectory(name=trajname, filename=filename, load_all=2,
                               dynamic_imports=SharedResult)

        with StorageContextManager(traj):
            a1 = traj.arrays.array
            self.assertTrue(a1[0, 0, 0] == 4.0)

            a2 = traj.arrays.carray
            self.assertTrue(a2[0, 1] == 4)

            a3 = traj.arrays.earray
            self.assertTrue(a3.read().shape == (3, 10, 3))

            a4 = traj.arrays.vlarray
            for idx, x in enumerate(a4):
                if idx == 0:
                    self.assertTrue(np.all(x == np.array(thevlarray)))
                elif idx == 1:
                    self.assertTrue(np.all(x == np.array(toappned)))
                else:
                    raise RuntimeError()

    def test_df(self):
        filename = make_temp_dir('hdf5errors.hdf5')
        traj = Trajectory(name=make_trajectory_name(self), filename=filename)
        traj.f_store()
        dadict = {'hi': [1, 2, 3, 4, 5], 'shu': ['bi', 'du', 'da', 'ha', 'hui']}
        dadict2 = {'answer': [42]}
        traj.f_add_result(SharedResult, 'dfs.df', SharedPandasFrame()).create_shared_data(data=pd.DataFrame(dadict))
        traj.f_add_result(SharedResult, 'dfs.df1', SharedPandasFrame()).create_shared_data(data=pd.DataFrame(dadict2))
        traj.f_add_result(SharedResult, 'dfs.df3', SharedPandasFrame())

        for irun in range(10):
            traj.df3.append(traj.df1.read())

        dframe = traj.df3.read()

        self.assertTrue(len(dframe) == 10)

        what = traj.df.select(where='index == 2')
        self.assertTrue(len(what) == 1)

    def test_errors(self):
        filename = make_temp_dir('hdf5errors.hdf5')
        traj = Trajectory(name=make_trajectory_name(self), filename=filename)

        npearray = np.ones((2, 10, 3), dtype=np.float)
        thevlarray = np.array(['j'.encode('utf-8'), 22.2, 'gutter'.encode('utf-8')])

        with self.assertRaises(TypeError):
            traj.f_add_result(SharedResult, 'arrays.vlarray', SharedVLArray()).create_shared_data(obj=thevlarray)
        traj.f_store()
        traj.arrays.vlarray.create_shared_data(obj=thevlarray)
        traj.f_add_result(SharedResult, 'arrays.array', SharedArray()).create_shared_data(data=npearray)
        traj.arrays.f_add_result(SharedResult, 'super.carray', SharedCArray(),
                                 comment='carray').create_shared_data(shape=(10, 10), atom=pt.atom.FloatAtom())
        traj.arrays.f_add_result(SharedResult, 'earray', SharedEArray()).create_shared_data('earray',
                                                                                            obj=npearray)

        traj.f_store()

        with self.assertRaises(TypeError):
            traj.arrays.array.iterrows()

        with StorageContextManager(traj):
            with self.assertRaises(RuntimeError):
                with StorageContextManager(traj):
                    pass
            self.assertTrue(traj.v_storage_service.is_open)
            with self.assertRaises(RuntimeError):
                StorageContextManager(traj).open_store()

        self.assertFalse(traj.v_storage_service.is_open)


class SharedTableTest(TrajectoryComparator):

    tags = 'unittest', 'trajectory', 'shared', 'hdf5', 'table', 'mehmet'

    def setUp(self):
        self.filename = make_temp_dir('shared_table_test.hdf5')

        self.traj = Trajectory(name=make_trajectory_name(self), filename=self.filename)

        self.traj.v_standard_result = SharedResult

        self.traj.f_store(only_init=True)

        self.traj.f_add_result('shared_data')

        self.shared_table = SharedTable(name='table',
                                        parent=self.traj.shared_data,
                                        trajectory=self.traj,
                                        add_to_parent=True)

    def test_table_read(self):
        the_reading_table = self.traj.results.shared_data.table
        self.assertTrue(the_reading_table is self.shared_table)
        the_reading_table.create_shared_data(description=MyTable)

        with StorageContextManager(self.traj):
            row = the_reading_table.row
            for i in range(10):
                row['id'] = i
                row['name'] = 'mehmet %d' % i
                row['surname'] = 'Timur'
                row['weight'] = 65.5 + i * 1.5
                row.append()
            the_reading_table.flush()

            for idx, row in enumerate(the_reading_table.iterrows()):
                self.assertEqual(row['id'], idx)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_reading_table = traj2.results.shared_data.table

        self.assertTrue(np.all(the_reading_table.read() == second_reading_table.read()))

        second_reading_table.append([(21, 'aaa', 'bbb', 100)])

        self.assertTrue(np.all(the_reading_table.read() == second_reading_table.read()))

        traj3 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        third_reading_table = traj3.results.shared_data.table

        self.assertTrue(np.all(the_reading_table.read() == third_reading_table.read()))

    def test_table_append(self):
        the_append_table = self.traj.results.shared_data.table
        self.assertTrue(the_append_table is self.shared_table)
        the_append_table.create_shared_data(description=MyTable)

        with StorageContextManager(self.traj):
            row = the_append_table.row
            for i in range(15):
                row['id'] = i * 2
                row['name'] = 'name %d' % i
                row['surname'] = '%d surname' % i
                row['weight'] = (i*0.5 + 50.0)
                row.append()
            the_append_table.flush()

            for idx, row in enumerate(the_append_table.iterrows()):
                self.assertEqual(row['id'], idx * 2)
                self.assertEqual(row['name'], ('name %d' % idx).encode('utf-8'))
                self.assertEqual(row['surname'], ('%d surname' % idx).encode('utf-8'))
                self.assertEqual(row['weight'], idx*0.5+50.0)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_append_table = traj2.results.shared_data.table

        with StorageContextManager(traj2):
            for idx, row in enumerate(second_append_table.iterrows()):
                self.assertEqual(row['id'], idx * 2)
                self.assertEqual(row['name'], ('name %d' % idx).encode('utf-8'))
                self.assertEqual(row['surname'], ('%d surname' % idx).encode('utf-8'))
                self.assertEqual(row['weight'], idx*0.5+50.0)

            second_append_table.append([(30, 'mehmet', 'timur', 65.5)])

            self.assertEqual(second_append_table.read(field='id')[-1], 30)
            self.assertEqual(second_append_table.read(field='name')[-1], 'mehmet'.encode('utf-8'))
            self.assertEqual(second_append_table.read(field='surname')[-1], 'timur'.encode('utf-8'))
            self.assertEqual(second_append_table.read(field='weight')[-1], 65.5)

        traj2.f_store()

        traj3 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        third_append_table = traj3.results.shared_data.table

        self.assertEqual((third_append_table.read(field='id')[-1]), 30)
        self.assertEqual((third_append_table.read(field='name')[-1]), 'mehmet'.encode('utf-8'))
        self.assertEqual((third_append_table.read(field='surname')[-1]), 'timur'.encode('utf-8'))
        self.assertEqual((third_append_table.read(field='weight')[-1]), 65.5)

        third_append_table.append([(33, 'Harrison', 'Ford', 95.5)])

        self.assertEqual((third_append_table.read(field='id')[-1]), 33)
        self.assertEqual((third_append_table.read(field='name')[-1]), 'Harrison'.encode('utf-8'))
        self.assertEqual((third_append_table.read(field='surname')[-1]), 'Ford'.encode('utf-8'))
        self.assertEqual((third_append_table.read(field='weight')[-1]), 95.5)

    def test_table_iterrows(self):
        the_iterrows_table = self.traj.results.shared_data.table
        self.assertTrue(the_iterrows_table is self.shared_table)
        the_iterrows_table.create_shared_data(description=MyTable)

        with StorageContextManager(self.traj):
            row = the_iterrows_table.row
            for i in range(10):
                row['id'] = i
                row['name'] = 'mehmet %d' % i
                row['surname'] = 'Timur'
                row['weight'] = 65.5 + i * 1.5
                row.append()
            the_iterrows_table.flush()

            for idx, row in enumerate(the_iterrows_table.iterrows()):
                self.assertEqual(row['id'], idx)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_iterrows_table = traj2.results.shared_data.table

        with StorageContextManager(traj2):
            for idx, row in enumerate(second_iterrows_table.iterrows()):
                self.assertEqual(row['id'], idx)

    def test_table_col(self):
        the_col_table = self.traj.results.shared_data.table

        self.assertTrue(the_col_table is self.shared_table)

        the_col_table.create_shared_data(description=MyTable)

        with StorageContextManager(self.traj):
            row = the_col_table.row
            for i in range(10):
                row['id'] = i
                row['name'] = 'mehmet %d' % i
                row['surname'] = 'Timur'
                row['weight'] = 65.5 + i * 1.5
                row.append()
            the_col_table.flush()

            for idx, row in enumerate(the_col_table.iterrows()):
                self.assertEqual(row['id'], idx)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_col_table = traj2.results.shared_data.table

        with StorageContextManager(traj2):
            for idx, row in enumerate(second_col_table.iterrows()):
                self.assertEqual(row['id'], idx)

        self.assertTrue(np.all(second_col_table.read(field='id') == second_col_table.col('id')))
        self.assertTrue(np.all(second_col_table.read(field='name') == second_col_table.col('name')))
        self.assertTrue(np.all(second_col_table.read(field='surname') == second_col_table.col('surname')))
        self.assertTrue(np.all(second_col_table.read(field='weight') == second_col_table.col('weight')))

    # def test_table_itersequence(self):
    #     pass
    #
    # def test_table_itersorted(self):
    #     pass
    #
    # def test_table_read_coordinates(self):
    #     pass
    #
    # def test_table_read_sorted(self):
    #     pass

    def test_table_getitem(self):
        the_getitem_table = self.traj.results.shared_data.table

        self.assertTrue(the_getitem_table is self.shared_table)

        the_getitem_table.create_shared_data(description=MyTable)

        with StorageContextManager(self.traj):
            row = the_getitem_table.row
            for i in range(10):
                row['id'] = i
                row['name'] = 'mehmet %d' % i
                row['surname'] = 'Timur'
                row['weight'] = 65.5 + i * 1.5
                row.append()
            the_getitem_table.flush()

            for idx, row in enumerate(the_getitem_table.iterrows()):
                self.assertEqual(row['id'], idx)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_getitem_table = traj2.results.shared_data.table

        with StorageContextManager(traj2):
            for idx, row in enumerate(second_getitem_table.iterrows()):
                self.assertTrue(np.all(second_getitem_table.read()[idx] == second_getitem_table[idx]))

            second_getitem_table.append([(30, 'mehmet nevvaf', 'timur', 65.5)])

            for idx, row in enumerate(second_getitem_table.iterrows(-1)):
                self.assertEqual(row['id'], 30)
                self.assertEqual(row['name'], 'mehmet nevvaf'.encode('utf-8'))
                self.assertEqual(row['surname'], 'timur'.encode('utf-8'))
                self.assertEqual(row['weight'], 65.5)

        traj2.f_store()

        traj3 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        third_getitem_table = traj3.results.shared_data.table

        with StorageContextManager(traj3):
            for idx, row in enumerate(third_getitem_table.iterrows()):
                self.assertTrue(np.all(third_getitem_table.read()[idx] == third_getitem_table[idx]))

    # def test_table_iter(self):
    #     pass
    #
    # def test_table_modify_column(self):
    #     pass
    #
    # def test_table_modify_columns(self):
    #     pass
    #
    # def test_table_modify_coordinates(self):
    #     pass
    #
    # def test_table_modify_rows(self):
    #     pass
    #
    # def test_table_remove_rows(self):
    #     pass
    #
    # def test_table_remove_row(self):
    #     pass

    def test_table_setitem(self):
        the_setitem_table = self.traj.results.shared_data.table

        self.assertTrue(the_setitem_table is self.shared_table)

        the_setitem_table.create_shared_data(description=MyTable)

        with StorageContextManager(self.traj):
            row = the_setitem_table.row
            for i in range(10):
                row['id'] = i
                row['name'] = 'mehmet %d' % i
                row['surname'] = 'Timur'
                row['weight'] = 65.5 + i * 1.5
                row.append()
            the_setitem_table.flush()

            for idx, row in enumerate(the_setitem_table.iterrows()):
                self.assertEqual(row['id'], idx)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_setitem_table = traj2.results.shared_data.table

        second_setitem_table[0] = [(100, 'Mehmet Nevvaf', 'TIMUR', 75.5)]

        self.assertEqual(second_setitem_table.read(field='id')[0], 100)
        self.assertEqual(second_setitem_table.read(field='name')[0], 'Mehmet Nevvaf'.encode('utf-8'))
        self.assertEqual(second_setitem_table.read(field='surname')[0], 'TIMUR'.encode('utf-8'))
        self.assertEqual(second_setitem_table.read(field='weight')[0], 75.5)

        traj2.f_store()

        traj3 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        third_setitem_table = traj3.results.shared_data.table

        self.assertEqual(third_setitem_table.read(field='id')[0], 100)
        self.assertEqual(third_setitem_table.read(field='name')[0], 'Mehmet Nevvaf'.encode('utf-8'))
        self.assertEqual(third_setitem_table.read(field='surname')[0], 'TIMUR'.encode('utf-8'))
        self.assertEqual(third_setitem_table.read(field='weight')[0], 75.5)

    # def test_table_get_where_list(self):
    #     pass
    #
    # def test_table_read_where(self):
    #     pass

    def test_table_where(self):
        the_where_table = self.traj.results.shared_data.table

        self.assertTrue(the_where_table is self.shared_table)

        the_where_table.create_shared_data(description=MyTable)

        with StorageContextManager(self.traj):
            row = the_where_table.row
            for i in range(10):
                row['id'] = i
                row['name'] = 'mehmet %d' % i
                row['surname'] = 'Timur'
                row['weight'] = 65.5 + i
                row.append()
            the_where_table.flush()

            for idx, row in enumerate(the_where_table.iterrows()):
                self.assertEqual(row['id'], idx)

            self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_where_table = traj2.results.shared_data.table

        with StorageContextManager(traj2):
            result = second_where_table.where('(id == 2)&(name == b"mehmet 2")&(surname ==b"Timur")&(weight == 67.5)')
            there = False
            for row in result:
                there = True
            self.assertTrue(there)

    # def test_table_append_where(self):
    #     pass
    #
    # def test_table_will_query_use_indexing(self):
    #     pass
    #
    # def test_table_copy(self):
    #     pass
    #
    # def test_table_flush_rows_to_index(self):
    #     pass
    #
    # def test_table_get_enum(self):
    #     pass
    #
    # def test_table_reindex(self):
    #     pass
    #
    # def test_table_reindex_dirty(self):
    #     pass
    #
    # def test_table_remove_index(self):
    #     pass
    #
    # def test_table_create_index(self):
    #     pass
    #
    # def test_table_create_cindex(self):
    #     pass
    #
    # def test_table_colindexes(self):
    #     pass
    #
    # def test_table_cols(self):
    #     pass
    #
    # def test_table_row(self):
    #     pass

    def test_table_flush(self):
        the_flush_table = self.traj.results.shared_data.table

        self.assertTrue(the_flush_table is self.shared_table)

        the_flush_table.create_shared_data(description=MyTable)

        with StorageContextManager(self.traj):
            row = the_flush_table.row
            for i in range(10):
                row['id'] = i
                row['name'] = 'mehmet %d' % i
                row['surname'] = 'Timur'
                row['weight'] = 65.5 + i
                row.append()
            the_flush_table.flush()

            for idx, row in enumerate(the_flush_table.iterrows()):
                self.assertEqual(row['id'], idx)
                self.assertEqual(row['name'], ('mehmet %d' % idx).encode('utf-8'))
                self.assertEqual(row['surname'], 'Timur'.encode('utf-8'))
                self.assertEqual(row['weight'], 65.5+idx)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_flush_table = traj2.results.shared_data.table

        with StorageContextManager(traj2):
            for idx, row in enumerate(second_flush_table.iterrows()):
                self.assertEqual(row['id'], idx)
                self.assertEqual(row['name'], ('mehmet %d' % idx).encode('utf-8'))
                self.assertEqual(row['surname'], 'Timur'.encode('utf-8'))
                self.assertEqual(row['weight'], 65.5+idx)

            row = second_flush_table.row
            for i in range(10, 11):
                row['id'] = i
                row['name'] = 'mehmet %d' % i
                row['surname'] = 'Timur'
                row['weight'] = 65.5 + i
                row.append()
            second_flush_table.flush()

            for idx, row in enumerate(second_flush_table.iterrows()):
                self.assertEqual(row['id'], idx)
                self.assertEqual(row['name'], ('mehmet %d' % idx).encode('utf-8'))
                self.assertEqual(row['surname'], 'Timur'.encode('utf-8'))
                self.assertEqual(row['weight'], 65.5+idx)


class SharedArrayTest(TrajectoryComparator):

    tags = 'unittest', 'trajectory', 'shared', 'hdf5', 'array', 'mehmet'

    def setUp(self):
        self.filename = make_temp_dir('shared_table_test.hdf5')

        self.traj = Trajectory(name=make_trajectory_name(self), filename=self.filename)

        self.traj.v_standard_result = SharedResult

        self.traj.f_store(only_init=True)

        self.traj.f_add_result('shared_data')

        self.shared_array = SharedArray(name='array',
                                        parent=self.traj.shared_data,
                                        trajectory=self.traj,
                                        add_to_parent=True)

    def test_array_read(self):
        the_reading_array = np.ones((100, 100)) * 4

        first_reading_array = self.traj.results.shared_data.array

        self.assertTrue(first_reading_array is self.shared_array)

        first_reading_array.create_shared_data(obj=the_reading_array)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_reading_array = traj2.shared_data.array.read()

        self.assertTrue(np.all(the_reading_array == second_reading_array),
                        '%s != %s' % (str(the_reading_array), str(second_reading_array)))

    def test_array_getitem(self):
        the_getitem_array = np.array(range(100))

        first_getitem_array = self.traj.results.shared_data.array

        first_getitem_array.create_shared_data(obj=the_getitem_array)

        for k in range(len(the_getitem_array)):
            self.assertEqual(the_getitem_array[k], first_getitem_array[k])

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        for j in range(len(the_getitem_array)):
            self.assertEqual(the_getitem_array[j], traj2.results.shared_data.array[j])

    def test_array_getenum(self):
        the_getenum_array = np.array(range(100))

        first_getenum_array = self.traj.results.shared_data.array

        first_getenum_array.create_shared_data(obj=the_getenum_array)

        with self.assertRaises(TypeError):
            first_getenum_array.get_enum()

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_enum_array = traj2.results.shared_data.array

        with self.assertRaises(TypeError):
            second_enum_array.get_enum()

    def test_array_iterrows(self):
        the_iterrows_array = np.random.randint(0, 100, (100, 100))

        first_iterrows_array = self.traj.results.shared_data.array

        first_iterrows_array.create_shared_data(obj=the_iterrows_array)

        with StorageContextManager(self.traj):
            for idx, row in enumerate(first_iterrows_array.iterrows()):
                self.assertTrue(np.all(row == the_iterrows_array[idx, :]))

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_iterrows_array = traj2.results.shared_data.array

        with StorageContextManager(traj2):
            for idx, row in enumerate(second_iterrows_array.iterrows()):
                self.assertTrue(np.all(row == the_iterrows_array[idx, :]))

    def test_array_setitem(self):
        the_setitem_array = np.zeros((50, 50))

        first_setitem_array = self.traj.results.shared_data.array

        first_setitem_array.create_shared_data(obj=the_setitem_array)

        first_setitem_array[2, 2] = 10

        self.assertEqual(first_setitem_array[2, 2], 10)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_setitem_array = traj2.results.shared_data.array

        self.assertEqual(second_setitem_array[2, 2], 10)

        second_setitem_array[3, 3] = 17

        self.assertEqual(second_setitem_array[3, 3], 17)

    def test_array_iter(self):

        the_iterrows_array = np.random.randint(0, 100, (100, 100))

        first_iterrows_array = self.traj.results.shared_data.array

        first_iterrows_array.create_shared_data(obj=the_iterrows_array)

        with StorageContextManager(self.traj):
            for idx, row in enumerate(first_iterrows_array):
                self.assertTrue(np.all(row == the_iterrows_array[idx, :]))

        self.assertTrue(np.all(the_iterrows_array == first_iterrows_array.read()))

        for idx, row in enumerate(the_iterrows_array):
            self.assertTrue(np.all(row == the_iterrows_array[idx, :]))

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_iterrows_array = traj2.results.shared_data.array

        with StorageContextManager(traj2):
            for idx, row in enumerate(second_iterrows_array):
                self.assertTrue(np.all(row == the_iterrows_array[idx, :]))

        self.assertTrue(np.all(the_iterrows_array == second_iterrows_array.read()))

        for idx, row in enumerate(second_iterrows_array):
            self.assertTrue(np.all(row == the_iterrows_array[idx, :]))

    def test_array_len(self):
        the_len_array = np.ones((100, 100))

        first_len_array = self.traj.results.shared_data.array

        self.assertTrue(first_len_array is self.shared_array)

        first_len_array.create_shared_data(obj=the_len_array)

        self.assertEqual(len(first_len_array), 100)

        self.traj.f_store()

        traj2 = load_trajectory(name=self.traj.v_name, filename=self.filename, load_all=2, dynamic_imports=SharedResult)

        second_len_array = traj2.results.shared_data.array

        self.assertEqual(len(second_len_array), 100)


if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)


"""This module contains constants defined for a global scale and used across most pypet modules.

It contains constants defining the maximum length of a parameter/result name or constants
that are recognized by storage services to determine how to store and load data.

"""

__author__ = 'Robert Meyer'


import sys

####################### VERSIONS_TO_STORE ####################
from pypet._version import __version__ as VERSION
import tables
tablesversion = tables.__version__
hdf5version = tables.hdf5_version
import pandas
pandasversion = pandas.__version__
import numpy
numpyversion = numpy.__version__
import scipy
scipyversion = scipy.__version__

try:
    import platform
    platformversion = ', '.join(platform.uname())
except Exception:
    platformversion = 'N/A'

try:
    import sumatra
    sumatraversion = sumatra.__version__
except ImportError:
    sumatraversion = 'N/A'

try:
    import dill
    dillversion = dill.__version__
except ImportError:
    dillversion = 'N/A'

try:
    import psutil
    psutilversion = psutil.__version__
except ImportError:
    psutilversion = 'N/A'

try:
    import git
    gitversion = git.__version__
except ImportError:
    gitversion = 'N/A'


python_version_string = '.'.join([str(x) for x in sys.version_info[0:3]])

VERSIONS_TO_STORE = {'pypet': VERSION, 'python': python_version_string,
                     'scipy': scipyversion, 'numpy': numpyversion, 'PyTables': tablesversion,
                     'pandas': pandasversion, 'Sumatra': sumatraversion,
                     'dill': dillversion, 'GitPython': gitversion,
                     'psutil': psutilversion, 'platform': platformversion,
                     'HDF5': hdf5version}


###################### Supported Data ########################

PARAMETERTYPEDICT = {bool.__name__: bool,
                     complex.__name__: complex,
                     float.__name__: float,
                     int.__name__: int,
                     numpy.bool_.__name__: numpy.bool_,
                     numpy.complex128.__name__: numpy.complex128,
                     numpy.complex64.__name__: numpy.complex64,
                     numpy.float32.__name__: numpy.float32,
                     numpy.float64.__name__: numpy.float64,
                     numpy.int16.__name__: numpy.int16,
                     numpy.int32.__name__: numpy.int32,
                     numpy.int64.__name__: numpy.int64,
                     numpy.int8.__name__: numpy.int8,
                     numpy.string_.__name__: numpy.string_,
                     numpy.uint16.__name__: numpy.uint16,
                     numpy.uint32.__name__: numpy.uint32,
                     numpy.uint64.__name__: numpy.uint64,
                     numpy.uint8.__name__: numpy.uint8,
                     str.__name__: str,
                     bytes.__name__: bytes}
""" A Mapping (dict) from the the string representation of a type and the type.

These are the so far supported types of the storage service and the standard parameter!
"""

# For compatibility with older pypet versions:
COMPATPARAMETERTYPEDICT = {}
for key in PARAMETERTYPEDICT.keys():
    dtype = PARAMETERTYPEDICT[key]
    typestr = repr(dtype)
    if 'class' in typestr:
        # Python 3 replaced "<type 'x'>" with "<class 'x'>"
        typestr=typestr.replace('class', 'type')

    COMPATPARAMETERTYPEDICT[typestr] = dtype

PARAMETER_SUPPORTED_DATA = (numpy.int8,
                            numpy.int16,
                            numpy.int32,
                            numpy.int64,
                            numpy.int,
                            numpy.int_,
                            numpy.long,
                            numpy.uint8,
                            numpy.uint16,
                            numpy.uint32,
                            numpy.uint64,
                            numpy.bool,
                            numpy.bool_,
                            numpy.float32,
                            numpy.float64,
                            numpy.float,
                            numpy.float_,
                            numpy.complex64,
                            numpy.complex,
                            numpy.complex_,
                            numpy.str_,
                            str,
                            bytes)
"""Set of supported scalar types by the storage service and the standard parameter"""



################### HDF5 Naming and Comments ########################

HDF5_STRCOL_MAX_NAME_LENGTH = 128
"""Maximum length of a (short) name"""
HDF5_STRCOL_MAX_LOCATION_LENGTH = 512
"""Maximum length of the location string"""
HDF5_STRCOL_MAX_VALUE_LENGTH = 64
"""Maximum length of a value string"""
HDF5_STRCOL_MAX_COMMENT_LENGTH = 512
"""Maximum length of a comment """
HDF5_STRCOL_MAX_RANGE_LENGTH = 1024
"""Maximum length of a parameter array summary """
HDF5_STRCOL_MAX_RUNTIME_LENGTH = 18
"""Maximum length of human readable runtime, 18 characters allows to display up to 999 days
excluding the microseconds"""
HDF5_MAX_OVERVIEW_TABLE_LENGTH = 1000
"""Maximum number of entries in an overview table"""


######## Multiprocessing Modes #############

WRAP_MODE_QUEUE = 'QUEUE'
"""For multiprocessing, queue multiprocessing mode """
WRAP_MODE_LOCK = 'LOCK'
""" Lock multiprocessing mode """
WRAP_MODE_NONE = 'NONE'
""" No multiprocessing wrapping for the storage service"""
WRAP_MODE_PIPE = 'PIPE'
"""Pipe multiprocessing mode"""
WRAP_MODE_LOCAL = 'LOCAL'
"""Data is only stored on the local machine"""
WRAP_MODE_NETLOCK = 'NETLOCK'
""" Lock multiprocessing mode over a network """
WRAP_MODE_NETQUEUE = 'NETQUEUE'
""" Queue multiprocessing mode over a network """


############ Loading Constants ###########################

LOAD_SKELETON = 1
"""For trajectory loading, loads only the skeleton."""
LOAD_DATA = 2
""" Loads skeleton and data."""
LOAD_NOTHING = 0
""" Loads nothing """
OVERWRITE_DATA = 3
"""Overwrites all data in RAM with data from disk"""
UPDATE_SKELETON = 1
""" DEPRECATED: Updates skeleton, i.e. adds only items that are not
part of your current trajectory."""
UPDATE_DATA = 2
""" DEPRECATED: Updates skeleton and data,
adds only items that are not part of your current trajectory."""


######### Storing Constants #####

STORE_NOTHING = 0
"""Stores nothing to disk"""
STORE_DATA_SKIPPING = 1
"""Stores only data of instances that have not been stored before"""
STORE_DATA = 2
"""Stored all data to disk adds to existing data"""
OVERWRITE_DATA = OVERWRITE_DATA
"""Overwrites data on disk"""


##################### STORING Message Constants ################################

LEAF = 'LEAF'
""" For trajectory or item storage, stores a *leaf* node, i.e. parameter or result object"""
TRAJECTORY = 'TRAJECTORY'
""" Stores the whole trajectory"""
MERGE = 'MERGE'
""" Merges two trajectories """
GROUP = 'GROUP'
""" Stores a group node, can be recursive."""
LIST = 'LIST'
""" Stores a list of different things, in order to avoid reopening and closing of the hdf5 file."""
SINGLE_RUN = 'SINGLE_RUN'
""" Stores a single run"""
PREPARE_MERGE = 'PREPARE_MERGE'
""" Updates a trajectory before it is going to be merged"""
BACKUP = 'BACKUP'
""" Backs up a trajectory"""
DELETE = 'DELETE'
""" Removes an item from hdf5 file"""
DELETE_LINK = 'DELETE_LINK'
""" Removes a soft link from hdf5 file"""
TREE = 'TREE'
""" Stores a subtree of the trajectory"""
ACCESS_DATA = 'ACCESS_DATA'
""" Access and manipulate data directly in the hdf5 file """
CLOSE_FILE = 'CLOSE_FILE'
""" Close a still opened HDF5 file """
OPEN_FILE = 'OPEN_FILE'
""" Opens an HDF5 file and keeps it open until `CLOSE_FILE` is passed. """
FLUSH = 'FLUSH'
""" Tells the storage to flush the file """


########## Names of Runs ####################

FORMAT_ZEROS = 8
""" Number of leading zeros"""
RUN_NAME = 'run_'
"""Name of a single run"""
RUN_NAME_DUMMY = 'run_ALL'
"""Dummy name if not created during run"""
FORMATTED_RUN_NAME = RUN_NAME + '%0' + str(FORMAT_ZEROS) + 'd'
"""Name formatted with leading zeros"""

SET_FORMAT_ZEROS = 5
""" Number of leading zeros for set"""
SET_NAME = 'run_set_'
"""Name of a run set"""
SET_NAME_DUMMY = 'run_set_ALL'
"""Dummy name if not created during run"""
FORMATTED_SET_NAME = SET_NAME + '%0' + str(SET_FORMAT_ZEROS) + 'd'
"""Name formatted with leading zeros"""


### Constants how to store individual leaf data into HDF5 ######

ARRAY = 'ARRAY'
"""Stored as array_

.. _array: http://pytables.github.io/usersguide/libref/homogenous_storage.html#the-array-class

"""
CARRAY = 'CARRAY'
"""Stored as carray_

.. _carray: http://pytables.github.io/usersguide/libref/homogenous_storage.html#the-carray-class

"""
EARRAY = 'EARRAY'
""" Stored as earray_e.

.. _earray: http://pytables.github.io/usersguide/libref/homogenous_storage.html#the-earray-class

"""

VLARRAY = 'VLARRAY'
"""Stored as vlarray_

.. _vlarray: http://pytables.github.io/usersguide/libref/homogenous_storage.html#the-vlarray-class

"""

TABLE = 'TABLE'
"""Stored as pytable_

.. _pytable: http://pytables.github.io/usersguide/libref/structured_storage.html#the-table-class

"""

DICT = 'DICT'
""" Stored as dict.

In fact, stored as pytable, but the dictionary wil be reconstructed.
"""

FRAME = 'FRAME'
""" Stored as pandas DataFrame_

.. _DataFrame: http://pandas.pydata.org/pandas-docs/dev/io.html#hdf5-pytables

"""

SERIES = 'SERIES'
""" Store data as pandas Series """

SPLIT_TABLE = 'SPLIT_TABLE'
""" If a table was split due to too many columns"""

DATATYPE_TABLE = 'DATATYPE_TABLE'
"""If a table contains the data types instead of the attrs"""

SHARED_DATA = 'SHARED_DATA_'
""" An HDF5 data object for direct interaction """

NESTED_GROUP = 'NESTED_GROUP'
""" An HDF5 group containing nested data """

############# LOGGING ############

LOG_ENV = '$env'
"""Wildcard replaced by name of environment"""
LOG_TRAJ = '$traj'
"""Wildcard replaced by name of trajectory"""
LOG_RUN = '$run'
"""Wildcard replaced by name of current run"""
LOG_PROC = '$proc'
"""Wildcard replaced by the name of the current process"""
LOG_HOST = '$host'
"""Wildcard replaced by the name of the current host"""
LOG_SET = '$set'
"""Wildcard replaced by the name of the current run set"""

DEFAULT_LOGGING = 'DEFAULT'
"""Default logging configuration"""


""" Module providing the functionality to allow automatic git commits.

* :func:`~pypet.gitintegration.make_git_commit` performs the git commit

* :func:`~pypet.gitintegration.add_commit_variables` adds some information about the commit to a
  :class:`~pypet.trajectory.Trajectory`.


"""

__author__ = 'Robert Meyer'

import time

try:
    import git
except ImportError:
    git = None

import pypet.pypetexceptions as pex


def add_commit_variables(traj, commit):
    """Adds commit information to the trajectory."""

    git_time_value = time.strftime('%Y_%m_%d_%Hh%Mm%Ss', time.localtime(commit.committed_date))

    git_short_name = str(commit.hexsha[0:7])
    git_commit_name = 'commit_%s_' % git_short_name
    git_commit_name = 'git.' + git_commit_name + git_time_value

    if not traj.f_contains('config.'+git_commit_name, shortcuts=False):

        git_commit_name += '.'
        # Add the hexsha
        traj.f_add_config(git_commit_name+'hexsha', commit.hexsha,
                          comment='SHA-1 hash of commit')

        # Add the description string
        traj.f_add_config(git_commit_name+'name_rev', commit.name_rev,
                          comment='String describing the commits hex sha based on '
                                  'the closest Reference')

        # Add unix epoch
        traj.f_add_config(git_commit_name+'committed_date',
                          commit.committed_date, comment='Date of commit as unix epoch seconds')

        # Add commit message
        traj.f_add_config(git_commit_name+'message', str(commit.message),
                          comment='The commit message')

        # # Add commit author
        # traj.f_add_config(git_commit_name+'committer', str(commit.committer.name),
        #                   comment='The committer of the commit')
        #
        # # Add author's email
        # traj.f_add_config(git_commit_name+'committer_email', str(commit.committer.email),
        #                   comment='Email of committer')


def make_git_commit(environment, git_repository, user_message, git_fail):
    """ Makes a commit and returns if a new commit was triggered and the SHA_1 code of the commit.

    If `git_fail` is `True` program fails instead of triggering a new commit given
    not committed changes. Then a `GitDiffError` is raised.

    """

    # Import GitPython, we do it here to allow also users not having GitPython installed
    # to use the normal environment

    # Open the repository
    repo = git.Repo(git_repository)
    index = repo.index

    traj = environment.v_trajectory

    # Create the commit message and append the trajectory name and comment
    if traj.v_comment:
        commentstr = ', Comment: `%s`' % traj.v_comment
    else:
        commentstr = ''

    if user_message:
        user_message += ' -- '

    message = '%sTrajectory: `%s`, Time: `%s`, %s' % \
              (user_message, traj.v_name, traj.v_time, commentstr)

    # Detect changes:
    diff = index.diff(None)

    if diff:
        if git_fail:
            # User requested fail instead of a new commit
            raise pex.GitDiffError('Found not committed changes!')
        # Make the commit
        repo.git.add('-u')
        commit = index.commit(message)
        new_commit = True

    else:
        # Take old commit
        commit = repo.commit(None)
        new_commit = False

    # Add the commit info to the trajectory
    add_commit_variables(traj, commit)

    return new_commit, commit.hexsha



__author__ = 'Henri Bunting'

import sys

import numpy as np
from pypet.parameter import BaseParameter

import unittest

try:
    import brian2
    from brian2.units.stdunits import mV, mA, kHz, ms
    from pypet.brian2.parameter import Brian2Parameter, Brian2Result, get_unit_fast
except ImportError:
    brian2 = None

from pypet.parameter import PickleParameter, ArrayParameter, SparseParameter
from pypet.tests.unittests.parameter_test import ParameterTest, ResultTest
from pypet.tests.testutils.ioutils import parse_args, run_suite
from pypet.utils.explore import cartesian_product

import logging
logging.basicConfig(level=logging.DEBUG)


@unittest.skipIf(brian2 is None, 'Can only be run with brian2!')
class Brian2ParameterTest(ParameterTest):

    tags = 'unittest', 'brian2', 'parameter', 'henri'

    def setUp(self):

        if not hasattr(self, 'data'):
            self.data = {}

        self.data['mV1'] = 42.0*mV
        self.data['ampere1'] = 1*mA
        self.data['integer'] = 16
        #self.data['kHz05'] = 0.5*kHz
        self.data['nested_array'] = np.array([[6.,7.,8.],[9.,10.,11.]]) * ms
        self.data['b2a'] = np.array([1., 2.]) * mV
        self.data['complex'] = np.array([1., 2.]) * mV*mV/mA**2.73

        super(Brian2ParameterTest, self).setUp()
        self.dynamic_imports = [Brian2Parameter]

    def make_params(self):
        self.param = {}
        for key, val in self.data.items():
            self.param[key] = Brian2Parameter(self.location+'.'+key, val, comment=key)

    def explore(self):
        self.explore_dict = cartesian_product({
                                            #'npstr': [np.array(['Uno', 'Dos', 'Tres']),
                                            #           np.array(['Cinco', 'Seis', 'Siette']),
                                            #           np.array(['Ocho', 'Nueve', 'Diez'])],
                                            'ampere1': [1*mA],
                                             #'val0': [1, 2, 3],
                                             'mV1': [42.0*mV, 3*mV, 4*mV],
                                             'b2a': [np.array([1., 2.]) * mV]})

        ## Explore the parameter:
        for key, vallist in self.explore_dict.items():
            self.param[key]._explore(vallist)
            self.assertTrue(self.param[key].v_explored and self.param[key].f_has_range())

    def test_supports(self):
        for key, val in self.data.items():
            self.assertTrue(self.param[key].f_supports(val))


    def test_false_on_values_not_of_same_type(self):
        self.assertFalse(self.param[list(self.param.keys())[0]]._values_of_same_type(11, 99*mV))


@unittest.skipIf(brian2 is None, 'Can only be run with brian2!')
class Brian2ParameterDuplicatesInStoreTest(unittest.TestCase):

    tags = 'unittest', 'brian2', 'parameter', 'store', 'henri'

    def setUp(self):
        self.data = {}
        self.data['brian2_single_a'] = 1. * mV
        self.data['brian2_array_b'] = np.array([3., 3., 4.]) * mV
        self.data['brian2_array_c'] = np.array([5.]) * mV
        self.data['brian2_array_d'] = np.array([[6.,7.,8.],[9.,10.,11.]]) * ms
        #self.data['brian2_mixedtype_array_a'] = np.array([9., 10.]) * mV
        self.data['brian2_mixedtype_array_b'] = np.array([13., 14.]) * mV

        self.location = 'MyName.Is.myParam'
        self.make_params()
        self.explore()

    def make_params(self):
        self.param = {}
        for key, val in self.data.items():
            self.param[key] = Brian2Parameter(self.location+'.'+key, val, comment=key)

    def explore(self):
        self.explore_dict = cartesian_product({#'brian2_array_a': [np.array([1., 2.]) * mV],
                                               # 'brian2_array_b': [2 * mV],
                                               # Arrays need to be of the same size!
                                               'brian2_array_c': [np.array([5., 8.]) * mV, np.array([7., 8.]) * mV],
                                               })


        ## Explore the parameter:
        for key, vallist in self.explore_dict.items():
            self.param[key]._explore(vallist)
            self.assertTrue(self.param[key].v_explored and self.param[key].f_has_range())


    def test_storage_and_loading(self):
        for key, param in self.param.items():
            store_dict = param._store()

            # Due to smart storing the storage dict should be small and only contain 5 items or less
            # 1 for data, 1 for reference, and 3 for the array/matrices/items
            if param.f_has_range():
                if isinstance(param,(ArrayParameter, PickleParameter)) and not isinstance(param, SparseParameter):
                    self.assertTrue(len(store_dict)<7)
                # For sparse parameter it is more:
                if isinstance(param, SparseParameter):
                    self.assertTrue(len(store_dict)<23)

            constructor = param.__class__

            param.f_unlock()
            param.f_empty()

            param = constructor('')

            param._load(store_dict)

            param._rename(self.location+'.'+key)

            self.param[key] = param

        self.test_the_insertion_made_implicetly_in_setUp()

        self.test_exploration()

        self.test_meta_settings()


    def test_the_insertion_made_implicetly_in_setUp(self):
        for key, val in self.data.items():
            if not key in self.explore_dict:
                self.param[key]._restore_default()
                param_val = self.param[key].f_get()
                self.assertTrue(np.all(repr(val) == repr(param_val)),'%s != %s'  %(str(val),str(param_val)))

    def test_exploration(self):
        for key, vallist in self.explore_dict.items():
            param = self.param[key]

            for idx, val in enumerate(vallist):
                assert isinstance(param, BaseParameter)
                param._set_parameter_access(idx)

                self.assertTrue(np.all(repr(param.f_get())==repr(val))),'%s != %s'%( str(param.f_get()),str(val))

                param_val = self.param[key].f_get_range()[idx]
                self.assertTrue(np.all(str(val) == str(param_val)),'%s != %s'  %(str(val),str(param_val)))

            param._restore_default()
            self.assertTrue(param.v_explored and param.f_has_range(), 'Error for %s' % key)
            val = self.data[key]
            self.assertTrue(np.all(repr(param.f_get())==repr(val))),'%s != %s'%( str(param.f_get()),str(val))

    def test_expanding(self):
        for key, vallist in self.explore_dict.items():
            param = self.param[key]

            copy_list = vallist.copy()
            old_len = len(vallist)

            param.f_unlock()
            param._expand(copy_list)

            new_len = len(param.f_get_range())

            self.assertEqual(new_len, 2 * old_len)

    def test_loading_and_expanding(self):
        # Regression test for issue #50
        # https://github.com/SmokinCaterpillar/pypet/issues/50
        for key, vallist in self.explore_dict.items():
            param = self.param[key]

            copy_list = vallist.copy()
            old_len = len(vallist)


            store_dict = param._store()
            param.f_unlock()
            param._load(store_dict)

            param.f_unlock()
            param._expand(copy_list)

            new_len = len(param.f_get_range())

            self.assertEqual(new_len, 2 * old_len)

    def test_meta_settings(self):
        for key, param in self.param.items():
            self.assertEqual(param.v_full_name, self.location+'.'+key)
            self.assertEqual(param.v_name, key)
            self.assertEqual(param.v_location, self.location)


@unittest.skipIf(brian2 is None, 'Can only be run with brian2!')
class Brian2ResultTest(ResultTest):

    tags = 'unittest', 'brian2', 'result', 'henri'

    def make_constructor(self):
        self.Constructor = Brian2Result
        self.dynamic_imports = [Brian2Result]

    def test_illegal_naming(self):
        for res in self.results.values():
            data_dict = {'val'+Brian2Result.IDENTIFIER:42}
            with self.assertRaises(AttributeError):
                res.f_set(**data_dict)


    def setUp(self):

        if not hasattr(self,'data'):
            self.data = {}

        self.data['mV1'] = 1*mV
        self.data['ampere1'] = 1*mA
        self.data['msecond17'] = 16*ms
        self.data['kHz05'] = 0.5*kHz

        self.data['mV_array'] = np.ones(20) * mV
        self.data['integer'] = 444
        self.data['complex'] = np.array([1., 2.]) * mV*mV/mA**-2.7343

        super(Brian2ResultTest, self).setUp()


@unittest.skipIf(brian2 is None, 'Can only be run with brian2!')
class Brian2GetUnitFastTest(unittest.TestCase):

    tags = 'unittest', 'brian2'

    def test_get_unit_fast(self):
        unit = get_unit_fast(42 * mV)
        self.assertEquals(unit, 1000 * mV)


if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)



__author__ = 'Robert Meyer'

import os
import unittest
import numpy as np
import tables
try:
    import dill
except ImportError:
    dill = None
import logging
import shutil

from pypet.trajectory import Trajectory
from pypet.utils.explore import cartesian_product
from pypet.environment import Environment
from pypet import pypetconstants
from pypet.parameter import Parameter
from pypet.tests.testutils.ioutils import run_suite, make_temp_dir, make_trajectory_name, \
     parse_args, get_log_config
from pypet.tests.testutils.data import create_param_dict, add_params, multiply, \
    simple_calculations, TrajectoryComparator, multiply_args


class CustomParameter(Parameter):

    def __init__(self, *args, **kwargs):
        super(CustomParameter, self).__init__(*args, **kwargs)


class Multiply(object):

    def __init__(self):
        self.var=42

    def __call__(self, traj, i):
        z = traj.x * traj.y + i
        traj.f_add_result('z', z)
        return z


@unittest.skipIf(dill is None, 'Only makes sense if dill is installed')
class ContinueTest(TrajectoryComparator):

    tags = 'integration', 'hdf5', 'environment', 'continue', 'dill'

    def tearDown(self):
        if hasattr(self, 'envs'):
            for env in self.envs:
                env.f_disable_logging()

        super(ContinueTest, self).tearDown()

    def make_run(self,env):

        ### Make a test run
        simple_arg = -13
        simple_kwarg= 13.0
        env.f_run(simple_calculations, simple_arg, simple_kwarg=simple_kwarg)

    def make_environment(self, idx, filename, continuable=True, delete_continue=False):

        #self.filename = '../../experiments/tests/HDF5/test.hdf5'
        self.logfolder = make_temp_dir(os.path.join('experiments',
                                                     'tests',
                                                     'Log'))
        self.cnt_folder = make_temp_dir(os.path.join('experiments','tests','cnt'))
        trajname = 'Test%d' % idx + '_' + make_trajectory_name(self)

        env = Environment(trajectory=trajname,
                          filename=filename,
                          file_title=trajname,
                          log_stdout=False,
                          log_config=get_log_config(),
                          continuable=continuable,
                          continue_folder=self.cnt_folder,
                          delete_continue=delete_continue,
                          large_overview_tables=True)


        self.envs.append(env)
        self.trajs.append( env.v_trajectory)


    def explore(self, traj):
        self.explored =cartesian_product({'Normal.trial': [0,1],
            'Numpy.double': [np.array([1.0,2.0,3.0,4.0]), np.array([-1.0,3.0,5.0,7.0])]})


        traj.f_explore(self.explored)

    # def _remove_nresults_from_traj(self, nresults):
    #
    #     hdf5_file = tables.open_file(self.filenames[0], mode='a')
    #     runtable = hdf5_file.getNode('/'+self.trajs[0].v_name +'/overview/runs')
    #
    #     for row in runtable.iterrows(0, nresults, 1):
    #         row['completed'] = 0
    #         row.update()
    #
    #     runtable.flush()
    #     hdf5_file.flush()
    #     hdf5_file.close()

    def _remove_nresults(self, traj, nresults, continue_folder):

        result_tuple_list = []

        n = 0
        for filename in os.listdir(continue_folder):
            _, ext = os.path.splitext(filename)

            if ext != '.rcnt':
                continue

            n += 1

            cnt_file = open(os.path.join(continue_folder, filename), 'rb')
            try:
                result = dill.load(cnt_file)
                cnt_file.close()
                result_tuple_list.append((result))
            except Exception:
                # delete broken files
                logging.getLogger().exception('Could not open continue snapshot '
                                              'file `%s`.' % filename)
                cnt_file.close()
                os.remove(filename)

        self.assertGreaterEqual(n, nresults)

        result_tuple_list = sorted(result_tuple_list, key=lambda x: x[0])
        timestamp_list = [x[1]['finish_timestamp'] for x in result_tuple_list]
        timestamp_list = timestamp_list[-nresults:]

        for timestamp in timestamp_list:
            filename = os.path.join(continue_folder, 'result_%s.rcnt' % repr(timestamp).replace('.','_'))
            os.remove(filename)

        result_tuple_list = []
        for filename in os.listdir(continue_folder):
            _, ext = os.path.splitext(filename)

            if ext != '.rcnt':
                continue

            cnt_file = open(os.path.join(continue_folder, filename), 'rb')
            result = dill.load(cnt_file)
            cnt_file.close()
            result_tuple_list.append((result))

        name_set = set([x[1]['name']  for x in result_tuple_list])
        removed = 0
        for run_name in traj.f_iter_runs():
            if run_name not in name_set:
                run_dict = traj.f_get_run_information(run_name, copy=False)
                run_dict['completed'] = 0
                idx = run_dict['idx']
                traj._updated_run_information.add(idx)
                removed += 1

        self.assertGreaterEqual(removed, nresults)
        logging.getLogger().error('Removed %s runs for continuing' % removed)
        traj.f_store(only_init=True)

    def test_continueing(self):
        self.filenames = [make_temp_dir('test_continuing.hdf5'), 0]

        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment( irun, filename, continuable=irun==1, delete_continue=False)

        self.param_dict={}
        create_param_dict(self.param_dict)

        for irun in range(len(self.filenames)):
            add_params(self.trajs[irun], self.param_dict)

        self.explore(self.trajs[0])
        self.explore(self.trajs[1])

        for irun in range(len(self.filenames)):
            self.make_run(self.envs[irun])

        traj_name = self.trajs[1].v_name
        continue_folder = os.path.join(self.cnt_folder, self.trajs[1].v_name)


        self.trajs[1]=self.envs[1].v_trajectory
        self._remove_nresults(self.trajs[1], 3, continue_folder)

        self.envs[1].v_current_idx = 0
        results = self.envs[1].resume(trajectory_name = traj_name)


        for irun in range(len(self.filenames)):
            self.trajs[irun].f_load(load_parameters=pypetconstants.OVERWRITE_DATA,
                                    load_derived_parameters=pypetconstants.OVERWRITE_DATA,
                                    load_results=pypetconstants.OVERWRITE_DATA,
                                    load_other_data=pypetconstants.OVERWRITE_DATA)


        self.compare_trajectories(self.trajs[0],self.trajs[1])
        #shutil.rmtree(self.cnt_folder)

        self.assertEqual(len(self.trajs[1]), len(results))

    def test_continueing_remove_completed(self):
        self.filenames = [make_temp_dir('test_continueing_remove_completed.hdf5')]

        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment( irun, filename, continuable=True, delete_continue=True)

        self.param_dict={}
        create_param_dict(self.param_dict)

        for irun in range(len(self.filenames)):
            add_params(self.trajs[irun], self.param_dict)


        self.explore(self.trajs[0])

        for irun in range(len(self.filenames)):
            self.make_run(self.envs[irun])

        traj_name = self.trajs[0].v_name
        continue_folder = os.path.join(self.cnt_folder, self.trajs[0].v_name)
        self.assertFalse(os.path.isdir(continue_folder))


    def test_removal(self):
        self.filenames = [make_temp_dir(os.path.join('experiments',
                                                      'tests',
                                                      'HDF5',
                                                      'removal.hdf5')), 0]

        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment( irun, filename)

        self.param_dict={}
        create_param_dict(self.param_dict)

        for irun in range(len(self.filenames)):
            add_params(self.trajs[irun], self.param_dict)


        self.explore(self.trajs[0])
        self.explore(self.trajs[1])


        for irun in range(len(self.filenames)):
            self.make_run(self.envs[irun])

        self.trajs[0].f_add_parameter('Delete.Me', 'I will be deleted!')
        self.trajs[0].f_store_item('Delete.Me')

        self.trajs[0].f_remove_item(self.trajs[0].f_get('Delete.Me'))

        self.assertTrue('Delete.Me' not in self.trajs[0],'Delete.Me is still in traj')

        self.trajs[0].f_load_skeleton()
        self.trajs[0].f_load_item('Delete.Me')
        self.trajs[0].f_delete_item(self.trajs[0].f_get('Delete.Me'),
                                        remove_from_trajectory=True)

        self.trajs[0].f_load_skeleton()
        self.assertTrue(not 'Delete.Me' in self.trajs[0],'Delete.Me is still in traj')


        for irun in range(len(self.filenames)):
            self.trajs[irun].f_load_child('results',recursive=True,load_data=pypetconstants.OVERWRITE_DATA)
            self.trajs[irun].f_load_child('derived_parameters',recursive=True,load_data=pypetconstants.OVERWRITE_DATA)

        self.compare_trajectories(self.trajs[0],self.trajs[1])


    def test_multiple_storage_and_loading(self):
        self.filenames = [make_temp_dir(os.path.join('experiments',
                                                      'tests',
                                                      'HDF5',
                                                      'multiple_storage_and_loading.hdf5')), 0]

        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment( irun, filename)

        self.param_dict={}
        create_param_dict(self.param_dict)

        for irun in range(len(self.filenames)):
            add_params(self.trajs[irun],self.param_dict)

        self.explore(self.trajs[0])
        self.explore(self.trajs[1])

        for irun in range(len(self.filenames)):
            self.make_run(self.envs[irun])

        #self.trajs[0].f_store()

        temp_sservice = self.trajs[0].v_storage_service
        temp_name = self.trajs[0].v_name

        self.trajs[0] = Trajectory()
        self.trajs[0].v_storage_service=temp_sservice
        self.trajs[0].f_load(name=temp_name, as_new=False, load_parameters=2,
                             load_derived_parameters=2, load_results=2, load_other_data=2)
        #self.trajs[0].f_load(trajectory_name=temp_name,as_new=False, load_params=2, load_derived_params=2, load_results=2)

        self.trajs[1].f_load_skeleton()
        self.trajs[1].f_load_items(self.trajs[1].f_to_dict().values(),only_empties=True)
        self.compare_trajectories(self.trajs[0],self.trajs[1])


@unittest.skipIf(dill is None, 'Only makes sense if dill is installed')
class ContinueMPTest(ContinueTest):

    # def test_removal(self):
    #     return super(ContinueMPTest, self).test_removal()

    tags = 'integration', 'hdf5', 'environment', 'continue', 'multiproc', 'nopool', 'dill'

    def make_run_mp(self,env):
        env.f_run(multiply)


    def make_environment(self, idx, filename, continuable=True, delete_continue=False,
                         add_time=True,
                         trajectory=None):
        #self.filename = '../../experiments/tests/HDF5/test.hdf5'
        self.logfolder = make_temp_dir(os.path.join('experiments',
                                                     'tests',
                                                     'Log'))
        self.cnt_folder = make_temp_dir(os.path.join('experiments',
                                                      'tests',
                                                      'cnt'))
        trajname = 'Test%d' % idx + '_' + make_trajectory_name(self)

        env = Environment(trajectory=trajname if trajectory is None else trajectory,
                          dynamically_imported_classes=[CustomParameter],
                          filename=filename,
                          add_time=add_time,
                          file_title=trajname,
                          log_stdout=False,
                          use_pool=False,
                          log_config=get_log_config(),
                          continuable=continuable,
                          continue_folder=self.cnt_folder,
                          delete_continue=delete_continue,
                          multiproc=True,
                          purge_duplicate_comments=False,
                          ncores=2)


        self.envs.append(env)
        self.trajs.append( env.v_trajectory)

    def explore_mp(self, traj):
        self.explored={'x':[0.0, 1.0, 2.0, 3.0, 4.0], 'y':[0.1, 2.2, 3.3, 4.4, 5.5]}

        traj.f_explore(cartesian_product(self.explored))

    def test_continueing_mp_custom(self):
        self.filenames = [make_temp_dir('test_continueing_mp_custom.hdf5'),
                          make_temp_dir('test_continueing_mp_custom2.hdf5')]

        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment(irun, filename, continuable=irun == 1)

        self.param_dict={'x':1.0, 'y':2.0}

        for irun in range(len(self.filenames)):
            self.trajs[irun].f_add_parameter(CustomParameter,'x', 1.0)
            self.trajs[irun].f_add_parameter(CustomParameter, 'y', 1.0)


        self.explore_mp(self.trajs[0])
        self.explore_mp(self.trajs[1])

        arg=33
        for irun in range(len(self.filenames)):

            self.envs[irun].f_run(Multiply(), arg)

        traj_name = self.trajs[1].v_name
        continue_folder = os.path.join(self.cnt_folder, self.trajs[1].v_name)

        self.envs.pop()
        self.assertEqual(len(self.envs), 1)
        self.make_environment(1, self.filenames[1], continuable=True, add_time=False,
                              trajectory=traj_name)
        self.trajs[1] = self.envs[1].v_traj
        self.trajs[1].f_load(load_data=pypetconstants.LOAD_NOTHING)
        self._remove_nresults(self.trajs[1], 3, continue_folder)

        results = self.envs[1].resume(trajectory_name = traj_name)
        results = sorted(results, key = lambda x: x[0])

        for irun in range(len(self.filenames)):
            self.trajs[irun].f_load(load_parameters=pypetconstants.OVERWRITE_DATA,
                                    load_derived_parameters=pypetconstants.OVERWRITE_DATA,
                                    load_results=pypetconstants.OVERWRITE_DATA,
                                    load_other_data=pypetconstants.OVERWRITE_DATA)

        self.compare_trajectories(self.trajs[1],self.trajs[0])

        for run_name in self.trajs[0].f_iter_runs():
            z = (self.trajs[0].v_idx, self.trajs[0].crun.z)
            self.assertTrue(z in results, '%s not in %s' % (z, results))

        self.assertTrue(len(self.trajs[-1])== len(results))
        #os.removedirs(self.cnt_folder)

    def test_continueing_mp(self):
        self.filenames = [make_temp_dir('test_continueing_mp.hdf5'), 0]

        self.envs=[]
        self.trajs = []

        for irun,filename in enumerate(self.filenames):
            if isinstance(filename,int):
                filename = self.filenames[filename]

            self.make_environment( irun, filename, continuable=irun==1, delete_continue=False)

        self.param_dict={'x':1.0, 'y':2.0}

        for irun in range(len(self.filenames)):
            self.trajs[irun].f_add_parameter(CustomParameter,'x', 1.0)
            self.trajs[irun].f_add_parameter(CustomParameter, 'y', 1.0)

        self.explore_mp(self.trajs[0])
        self.explore_mp(self.trajs[1])

        for irun in range(len(self.filenames)):
            self.envs[irun].f_run(multiply)

        traj_name = self.trajs[1].v_name
        continue_folder = os.path.join(self.cnt_folder, self.trajs[1].v_name)


        self.trajs[1]=self.envs[1].v_trajectory
        self._remove_nresults(self.trajs[1], 3, continue_folder)

        self.envs[1].v_current_idx = 0
        results = self.envs[1].resume(trajectory_name = traj_name)
        results = sorted(results, key = lambda x: x[0])

        for irun in range(len(self.filenames)):
            self.trajs[irun].f_load(load_parameters=pypetconstants.OVERWRITE_DATA,
                                    load_derived_parameters=pypetconstants.OVERWRITE_DATA,
                                    load_results=pypetconstants.OVERWRITE_DATA,
                                    load_other_data=pypetconstants.OVERWRITE_DATA)

        self.compare_trajectories(self.trajs[1],self.trajs[0])

        self.assertEqual(len(self.trajs[1]), len(results))

        for run_name in self.trajs[0].f_iter_runs():
            z = (self.trajs[0].v_idx, self.trajs[0].crun.z)
            self.assertTrue( z in results, '%s not in %s' % (z, results))


@unittest.skipIf(dill is None, 'Only makes sense if dill is installed')
class ContinueMPPoolTest(ContinueMPTest):

    tags = 'integration', 'hdf5', 'environment', 'continue', 'multiproc', 'pool', 'dill'

    def make_environment_mp(self, idx, filename):
        #self.filename = '../../experiments/tests/HDF5/test.hdf5'
        self.logfolder = make_temp_dir(os.path.join('experiments',
                                                      'tests',
                                                      'Log'))
        self.cnt_folder = make_temp_dir(os.path.join('experiments',
                                                      'tests',
                                                      'cnt'))
        trajname = 'Test%d' % idx + '_' + make_trajectory_name(self)

        env = Environment(trajectory=trajname,
                          dynamic_imports=[CustomParameter],
                          filename=filename,
                          file_title=trajname,
                          log_stdout=False,
                          purge_duplicate_comments=False,
                          log_config=get_log_config(),
                          continuable=True,
                          continue_folder=self.cnt_folder,
                          delete_continue=False,
                          multiproc=True,
                          use_pool=True,
                          ncores=4)

        self.envs.append(env)
        self.trajs.append( env.v_trajectory)


if __name__ == '__main__':
    opt_args = parse_args()
    run_suite(**opt_args)